#include <SPI.h>
#include <Gamebuino.h>

extern const byte font3x5[];

extern const byte warps[];

extern const byte* events[];
extern const byte PROGMEM house_events[];
extern const byte PROGMEM town_events[];
extern const byte PROGMEM forest_events[];

extern const byte* const PROGMEM dialogs[];
extern const byte PROGMEM mom_dialog[];
extern const byte PROGMEM town_sign[];

extern const byte* maps[];
extern const byte PROGMEM map_house[];
extern const byte PROGMEM map_town[];
extern const byte PROGMEM map_forest[];

extern const byte tiles[];
extern const byte* tiles_pointer;
extern const byte player_charset[];
extern const byte* player_charset_pointer;
extern const byte PROGMEM player_mask01[];
extern const byte PROGMEM player_mask02[];
extern const byte PROGMEM player_mask03[];
extern const byte PROGMEM player_mask04[];
extern const byte PROGMEM player_mask05[];
extern const byte PROGMEM player_mask06[];
extern const byte* player_masks_pointer[];
extern const byte PROGMEM slime_01_sprite[];
extern const byte PROGMEM slime_02_sprite[];
extern const byte PROGMEM slime_03_sprite[];
extern const byte PROGMEM slime_04_sprite[];
extern const byte PROGMEM slime_05_sprite[];
extern const byte PROGMEM slime_06_sprite[];
extern const byte PROGMEM slime_07_sprite[];
extern const byte PROGMEM slime_08_sprite[];
extern const byte PROGMEM slime_09_sprite[];
extern const byte PROGMEM slime_10_sprite[];
extern const byte* slimes_pointer[];

extern const byte* const PROGMEM slimes_names[];
extern const byte PROGMEM slime_01_name[];
extern const byte PROGMEM slime_02_name[];
extern const byte PROGMEM slime_03_name[];
extern const byte PROGMEM slime_04_name[];
extern const byte PROGMEM slime_05_name[];
extern const byte PROGMEM slime_06_name[];
extern const byte PROGMEM slime_07_name[];
extern const byte PROGMEM slime_08_name[];
extern const byte PROGMEM slime_09_name[];
extern const byte PROGMEM slime_10_name[];
extern const byte PROGMEM slimes_stats[];
extern const byte player_current_slimes[];
extern const byte player_slimes_data[];

#define TILE_WIDTH 8
#define TILE_HEIGHT 8
#define TILES_PASSABLE_END 11
#define TILES_ANIMATED_START 9
#define TILES_ANIMATED_END 11
#define TILE_BATTLE 3
#define ANIMATION_FREQUENCY 500 // ms
#define KEY_DELAY 5
#define EVENT_WARP 1
#define EVENT_LOAD_MAP 2
#define EVENT_DIALOG 3

#define PLAYER_SLIMES_DATA_LENGTH 14
#define EVENTS_LENGTH 4

#define TYPE_GRASS 0
#define TYPE_FIRE 1
#define TYPE_WATER 2
#define TYPE_NORMAL 3
#define TYPE_POISON 4
#define TYPE_ROCK 5
#define TYPE_GOLD 6

Gamebuino gb;

//          gb.display.cursorX = 0;
//          gb.display.cursorY = 0;
//          gb.display.print(house_events[0]);
//          gb.display.update();
//          delay(5000);

void setup(){
  gb.begin();
  gb.display.setFont(font3x5);
  gb.display.persistence = true;
  byte player_x = 8;
  byte player_y = 1;
  byte player_direction = 3;
  byte player_animation = 0;
  const byte* current_map = map_house;
  const byte* current_events = house_events;
  byte event = EVENT_LOAD_MAP;
  byte event_id;
  byte map_width;
  byte map_height;
  int camera_x;
  int camera_y;
  int key_delay = 0;
  while(1){
    if(gb.update()){
      if(event){
        switch(event){
          case EVENT_WARP :
            gb.display.clear();
            gb.display.update();
            current_map = maps[warps[pgm_read_byte(current_events+event_id+3)*EVENTS_LENGTH]];
            player_x = warps[pgm_read_byte(current_events+event_id+3)*EVENTS_LENGTH+1];
            player_y = warps[pgm_read_byte(current_events+event_id+3)*EVENTS_LENGTH+2];
            player_direction = warps[pgm_read_byte(current_events+event_id+3)*EVENTS_LENGTH+3];
            current_events = events[warps[pgm_read_byte(current_events+event_id+3)*EVENTS_LENGTH]];
          case EVENT_LOAD_MAP :
            map_width = pgm_read_byte(current_map);
            map_height = pgm_read_byte(current_map+1);
            camera_x = player_x*TILE_WIDTH-LCDWIDTH/2+4;
            camera_x = camera_x*(camera_x > 0)+(map_width*TILE_WIDTH-LCDWIDTH-camera_x)*(camera_x > map_width*TILE_WIDTH-LCDWIDTH);
            camera_y = player_y*TILE_HEIGHT-LCDHEIGHT/2+4;
            camera_y = camera_y*(camera_y > 0)+(map_height*TILE_HEIGHT-LCDHEIGHT-camera_y)*(camera_y > map_height*TILE_HEIGHT-LCDHEIGHT);
            event = 0;
            break;
          case EVENT_DIALOG :
            gb.display.setColor(BLACK);
            gb.display.fillRect(0, 21, 84, 27);
            gb.display.setColor(WHITE);
            gb.display.fillRect(1, 22, 82, 25);
            gb.display.setColor(BLACK);
            gb.display.cursorX = 2;
            gb.display.cursorY = 23;
            gb.display.print((const __FlashStringHelper*)pgm_read_word(&dialogs[event_id]));
            gb.display.cursorX = 79;
            gb.display.cursorY = 40;
            gb.display.print("\25");
            gb.display.update();
            wait_for_key(BTN_A);
            key_delay = KEY_DELAY;
          default :
            event = 0;
        }
      }
      gb.display.clear();
      draw_map(current_map, camera_x, camera_y);
      draw_player(player_x*TILE_WIDTH-camera_x, player_y*TILE_WIDTH-camera_y, player_direction, player_animation);
      char x_temp = -gb.buttons.repeat(BTN_LEFT, 1)*(player_x > 0)+gb.buttons.repeat(BTN_RIGHT, 1)*(player_x < map_width-1);
      char y_temp = -gb.buttons.repeat(BTN_UP, 1)*(player_y > 0)+gb.buttons.repeat(BTN_DOWN, 1)*(player_y < map_height-1);
      if(!x_temp != !y_temp){
        player_direction = (1+x_temp)*(x_temp != 0);
        player_direction = (2+y_temp)*(y_temp != 0 || player_direction == 0);
        if(TILES_PASSABLE_END-pgm_read_byte(current_map+2+(player_y+y_temp)*map_width+player_x+x_temp) > 0){
          for(byte i = 1; i <= 8; i++){
            player_animation += 1*i%2;
            player_animation *= (player_animation < 3 && i < 8);
            camera_x = (player_x*TILE_WIDTH-LCDWIDTH/2+4+i*x_temp);
            camera_x = camera_x*(camera_x > 0)+(map_width*TILE_WIDTH-LCDWIDTH-camera_x)*(camera_x > map_width*TILE_WIDTH-LCDWIDTH);
            camera_y = player_y*TILE_HEIGHT-LCDHEIGHT/2+4+i*y_temp;
            camera_y = camera_y*(camera_y > 0)+(map_height*TILE_HEIGHT-LCDHEIGHT-camera_y)*(camera_y > map_height*TILE_HEIGHT-LCDHEIGHT);
            gb.display.clear();
            draw_map(current_map, camera_x, camera_y);
            draw_player(player_x*TILE_WIDTH-camera_x+i*x_temp, player_y*TILE_WIDTH-camera_y+i*y_temp, player_direction, player_animation);
            gb.display.update();
            delay(10);
          }
          player_x += x_temp;
          player_y += y_temp;
          if(pgm_read_byte(current_map+2+(player_y)*map_width+player_x) == TILE_BATTLE){

            // WILD BATTLE

            // draw battle screen
            gb.display.clear();
            draw_battle_box(0, 6);
            draw_battle_box(32, 39);

            // load player's slime
            byte player_current_slime_id = player_slimes_data[(player_current_slimes[0]-1)*PLAYER_SLIMES_DATA_LENGTH]-1;
            gb.display.drawBitmap(0, 16, slimes_pointer[player_current_slime_id]);
            gb.display.cursorX = 32;
            gb.display.cursorY = 33;
            gb.display.print((const __FlashStringHelper*)pgm_read_word(&slimes_names[player_current_slime_id]));
            gb.display.update();
            wait_for_key(BTN_B);

            //
            
          }
          for(byte i = 1; i <= pgm_read_byte(current_events)*EVENTS_LENGTH; i+=EVENTS_LENGTH){
            if(player_x == pgm_read_byte(current_events+i) && player_y == pgm_read_byte(current_events+i+1)){
              event = pgm_read_byte(current_events+i+2);
              event_id = i;
            }
          }
        }
      }else if(gb.buttons.pressed(BTN_A)*(key_delay == 0)){
        for(byte i = 1; i <= pgm_read_byte(current_events)*4; i+=4){
          if(player_x-(player_direction == 2)+(player_direction == 0) == pgm_read_byte(current_events+i) && player_y-(player_direction == 1)+(player_direction == 3) == pgm_read_byte(current_events+i+1)){
            event = pgm_read_byte(current_events+i+2);
            event_id = pgm_read_byte(current_events+i+3);
          }
        }
      }
      key_delay -= 1*(key_delay > 0);
    }
  }
}

void loop(){
}

void draw_map(const byte map_[], int camera_x, int camera_y){
  byte map_width = pgm_read_byte(map_);
  byte map_height = pgm_read_byte(map_+1);
  for(byte y = 0; y <= 6; y++){
    for(byte x = 0; x <= 11; x++){
      int tile_x = camera_x/TILE_WIDTH+x;
      int tile_y = camera_y/TILE_HEIGHT+y;
      if(tile_x >= 0 && tile_x < map_width && tile_y >= 0 && tile_y < map_height){
        byte tile_num = pgm_read_byte(map_+2+tile_y*map_width+tile_x);
        tile_num += (tile_num >= TILES_ANIMATED_START && tile_num <= TILES_ANIMATED_END)*millis()/ANIMATION_FREQUENCY%2;
        draw_sprite(tiles_pointer+tile_num*TILE_HEIGHT, x*TILE_WIDTH-camera_x%TILE_WIDTH, y*TILE_HEIGHT-camera_y%TILE_HEIGHT);
      }
    }
  }
}

void draw_player(int x, int y, byte direction, byte animation){
  gb.display.setColor(WHITE, BLACK);
  gb.display.drawBitmap(x, y, player_masks_pointer[direction*3+animation]);
  draw_sprite(player_charset_pointer+(direction*24+animation*8), x, y);
}


void draw_sprite(const byte data[], char x, char y){  // routine by Sorunome
  uint8_t* buf = (((y+8)&0xF8)>>1) * 21 + x + gb.display.getBuffer();
  asm volatile(
  "mov R20,%[y]\n\t"
  "ldi R17,7\n\t"
  "add R20,R17\n\t"
  "brmi End\n\t"
  "cpi %[y],48\n\t"
  "brpl End\n\t"
  "inc R20\n\t"
  "ldi R16,8\n\t"
  "andi R20,7\n\t"
  "cpi R20,0\n\t"
  "breq LoopAligned\n"
  "LoopStart:\n\t"
  "tst %[x]\n\t"
  "brmi LoopSkip\n\t"
  "cpi %[x],84\n\t"
  "brcc LoopSkip\n\t"
  "ld R17,Z\n\t"
  "eor R18,R18\n\t"
  "mov R19,R20\n\t"
  "clc\n\t"
  "LoopShift:\n\t" // carry is still reset from the cpi instruction or from the dec
  "rol R17\n\t"
  "rol R18\n\t"
  "dec R19\n\t"
  "brne LoopShift\n\t"
  "tst %[y]\n\t"
  "brmi LoopSkipPart\n\t"
  "ld R19,X\n\t"
  "or R19,R17\n\t"
  "st X,R19\n\t"
  "LoopSkipPart:\n\t"
  "cpi %[y],48\n\t"
  "brpl LoopSkip\n\t"
  "ld R19,Y\n\t"
  "or R19,R18\n\t"
  "st Y,R19\n\t"
  "LoopSkip:\n\t"
  "eor R18,R18\n\t"
  "ldi R19,1\n\t"
  "add R26,R19\n\t" // INC DOESN'T CHANGE CARRY!
  "adc R27,R18\n\t"
  "add R28,R19\n\t"
  "adc R29,R18\n\t"
  "add R30,R19\n\t"
  "adc R31,R18\n\t"
  "inc %[x]\n\t"
  "dec R16\n\t"
  "brne LoopStart\n\t"
  "rjmp End\n"
  "LoopAligned:\n\t"
  "tst %[x]\n\t"
  "brmi LoopAlignSkip\n\t"
  "cpi %[x],84\n\t"
  "brcc LoopAlignSkip\n\t"
  "ld R17,Z\n\t"
  "ld R18,X\n\t"
  "eor R18,R17\n\t"
  "st X,R18\n\t"
  "LoopAlignSkip:\n\t"
  "ldi R18,1\n\t"
  "add R26,R18\n\t"
  "adc R27,R20\n\t"
  "add R30,R18\n\t"
  "adc R31,R20\n\t"
  "inc %[x]\n\t"
  "dec R16\n\t"
  "brne LoopAligned\n"
  "End:\n\t"
  ::"x" (buf - 84),"y" (buf),"z" (data),[y] "r" (y),[x] "r" (x):"r16","r17","r18","r19","r20");
}

void wait_for_key(int key){
  while(!gb.buttons.pressed(key)){
    delay(10);
    gb.update();
  }
}

void draw_battle_box(int x, int y){
  gb.display.setColor(BLACK);
  gb.display.drawFastHLine(x+1, y, 50);
  gb.display.drawFastHLine(x, y+1, 52);
  gb.display.drawFastHLine(x+1, y+2, 50);
  gb.display.setColor(WHITE);
  gb.display.drawFastHLine(x+1, y+1, 50);
  gb.display.setColor(BLACK);
}
